;; Generated package description from phpinspect.el  -*- no-byte-compile: t -*-
(define-package "phpinspect" "2.0.1.0.20240910.210823" "PHP parsing and code intelligence package" '((compat "29")) :commit "4041796ee00524c1db720a061c41d8c96149bf81" :authors '(("Hugo Thunnissen" . "devel@hugot.nl")) :maintainer '("Hugo Thunnissen" . "devel@hugot.nl") :keywords '("php" "languages" "tools" "convenience") :url "https://github.com/hugot/phpinspect.el")
